"""Example of using Physalia to profile F-Droid.

In this script free top apps for music streaming were profiled
in terms of energy consumption in common use cases.
"""

from time import sleep
from samples.utils import AndroidViewClientUseCase

# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# this is just a script -- uppercase variables would look odd

PLAY_DURATION = 30 # seconds

def cleanup(use_case):
    """Close the app and wait a second."""
    use_case.uninstall_app()
    sleep(1)

def run(_):
    sleep(PLAY_DURATION)

# Spotify
def prepare_spotify(use_case):
    use_case.prepare_apk()
    use_case.open_app()
    use_case.wait_for_text("Log in").touch()
    use_case.wait_for_text("Log in with Facebook").touch()
    use_case.wait_for_text("Home")
    use_case.wait_for_text("Your Music").touch()
    use_case.wait_for_text("Songs").touch()
    use_case.wait_for_text("SHUFFLE PLAY").touch()

spotify_use_case = AndroidViewClientUseCase(
    "PlaySong",
    "./apks/Spotify Music_v7.5.0.1076_apkpure.com.apk",
    "com.spotify.music",
    "7.5.0.1076",
    run,
    prepare=prepare_spotify,
    cleanup=cleanup
)

# spotify_use_case.profile_and_persist()

# SongFlip - Free Music & Player
# com.hypermedia.songflip

def prepare_songflip(use_case):
    use_case.prepare_apk()
    use_case.open_app()
    use_case.wait_for_text("Start").touch()
    use_case.wait_for_id("com.hypermedia.songflip:id/song_title").touch()
    use_case.wait_for_id("com.hypermedia.songflip:id/song_artist").touch()

songflip_use_case = AndroidViewClientUseCase(
    "PlaySong",
    "./apks/SongFlip Free Music Player_v1.1.8_apkpure.com.apk",
    "com.hypermedia.songflip",
    "1.1.8",
    run,
    prepare=prepare_songflip,
    cleanup=cleanup
)

songflip_use_case.profile_and_persist()

# SoundCloud

def prepare_soundcloud(use_case):
    use_case.prepare_apk()
    use_case.open_app()
    use_case.wait_for_text("Sign in").touch()
    use_case.wait_for_text("Sign in with Facebook").touch()
    use_case.wait_for_text("Continue").touch()
    use_case.wait_for_id(
        "com.soundcloud.android:id/tab_layout"
    ).children[2].touch()
    use_case.wait_for_text("Liked tracks").touch()
    use_case.wait_for_id("com.soundcloud.android:id/track_list_item").touch()

soundcloud_use_case = AndroidViewClientUseCase(
    "PlaySong",
    "./apks/SoundCloud Music Audio_v2017.01.24-release_apkpure.com.apk",
    "com.soundcloud.android",
    "2017.01.24-release",
    run,
    prepare=prepare_soundcloud,
    cleanup=cleanup
)

soundcloud_use_case.profile_and_persist()

# Deezer
# deezer.android.app

def prepare_deezer(use_case):
    use_case.prepare_apk()
    use_case.open_app()
    use_case.wait_for_text("Log in").touch()
    use_case.wait_for_text("Log in with email address").touch()
    use_case.wait_for_text("Facebook").touch()
    use_case.wait_for_content_description("Not Now ").touch()
    use_case.wait_for_id("deezer.android.app:id/mosaic_cover_subtitle").touch()

deezer_use_case = AndroidViewClientUseCase(
    "PlaySong",
    "./apks/Deezer Music Song Streaming_v5.4.0.88_apkpure.com.apk",
    "deezer.android.app",
    "5.4.0.88",
    run,
    prepare=prepare_deezer,
    cleanup=cleanup
)

deezer_use_case.profile_and_persist()

# YOUZEEK
# com.youzeek.AndroidApp

def prepare_youzeek(use_case):
    use_case.prepare_apk()
    use_case.open_app()
    use_case.wait_for_id("com.youzeek.AndroidApp:id/login_facebook").touch()
    use_case.wait_for_content_description("Interstitial close button").touch()
    use_case.wait_for_text("Next").touch()
    use_case.wait_for_text("Next").touch()
    use_case.wait_for_text("Next").touch()
    use_case.wait_for_text("Maybe later").touch()
    use_case.wait_for_id("com.youzeek.AndroidApp:id/songTitle").touch()

youzeek_use_case = AndroidViewClientUseCase(
    "PlaySong",
    "./apks/YOUZEEK Free Music Streaming_v4.4.6_apkpure.com.apk",
    "com.youzeek.AndroidApp",
    "4.4.6",
    run,
    prepare=prepare_youzeek,
    cleanup=cleanup
)

youzeek_use_case.profile_and_persist()

# Free music for YouTube: Stream
# com.djit.apps.stream
# FIXME this is an app that reproduces youtube videos - SKIP IT

# Panasonic Music Streaming
# com.panasonic.avc.diga.wwmusicstreaming
# Skipped because the purpose of this app is to stream music to devices
# in the network


# Musicsense - Music Streaming
# com.tentracks.musicsense

def prepare_musicsense(use_case):
    use_case.prepare_apk()
    use_case.open_app()
    use_case.wait_for_text("Allow").touch()
    use_case.wait_for_text("Allow").touch()
    use_case.wait_for_text("LOG IN").touch()
    use_case.wait_for_text("START LISTENING").touch()

musicsense_use_case = AndroidViewClientUseCase(
    "PlaySong",
    "./apks/Musicsense Music Streaming_v1.1.15_apkpure.com.apk",
    "com.tentracks.musicsense",
    "1.1.15",
    run,
    prepare=prepare_musicsense,
    cleanup=cleanup
)

musicsense_use_case.profile_and_persist()

# Napster
# com.rhapsody.napster
# Could not find a freemium version

# TIDAL
# com.aspiro.tidal
# Does not provide a free version (only tria)

# Apple Music
# com.apple.android.music
# Does not have a free streaming service



"""Models that require persistence."""

import csv
import os
from pathlib import Path

from itertools import groupby
from collections import OrderedDict
from operator import itemgetter
import bisect
import numpy

class Measurement(object):
    """Energy measurement information.

    Attributes:
        timestamp               When the execution started.
        use_case                Key identifier of the use case.
        app_pkg                 Package of the app.
        app_version             Version of the app.
        device_model            Device where the measurements were performed.
        duration                Time it takes to execute the use case.
        energy_consumption      Mean of the measurements.
        power_meter             Name of the power meter used.

    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # Eight is reasonable in this case.

    csv_storage = "./db.csv"
    COLUMN_APP_PKG = 2
    COLUMN_USE_CASE = 1
    COLUMN_NAME = COLUMN_USE_CASE

    def __init__(
            self,
            timestamp,
            use_case,
            app_pkg,
            app_version,
            device_model,
            duration,
            energy_consumption,
            power_meter="NA",
            success=True,
            notes=None,
    ):  # noqa: D102,D107
        self.persisted = False
        self.timestamp = float(timestamp)
        self.use_case = use_case
        self.app_pkg = app_pkg
        self.app_version = app_version
        self.device_model = device_model
        self.duration = float(duration)
        self.energy_consumption = float(energy_consumption)
        self.power_meter = power_meter
        self.success = success
        self.notes = notes

    def persist(self):
        """Store measurement in the database."""
        if self.persisted:
            return False
        self.save_to_csv(self.csv_storage)
        self.persisted = True
        return True

    def save_to_csv(self, filename):
        """Store measurements in a CSV file."""
        if not Path(filename).is_file():
            with open(filename, 'wt') as csvfile:
                csv_writer = csv.writer(csvfile)
                # csv_writer.writeheader()
                csv_writer.writerow([
                    "timestamp",
                    "use_case",
                    "app_pkg",
                    "app_version",
                    "device_model",
                    "duration",
                    "energy_consumption",
                    "power_meter",
                    "success",
                    "notes",
                ])
        with open(filename, 'at') as csvfile:
            csv_writer = csv.writer(csvfile)
            csv_writer.writerow([
                self.timestamp,
                self.use_case,
                self.app_pkg,
                self.app_version,
                self.device_model,
                self.duration,
                self.energy_consumption,
                self.power_meter,
                self.success,
                self.notes,
            ])

    def __str__(self):
        """Get description of the measurement."""
        return (
            "Measurement for {}:\n"
            "  {: <20}{:.4f}J\n"
            "  {: <20}{}s\n"
            "  {: <20}{}\n"
            "  {: <20}{}"
        ).format(self.use_case,
                 "Energy consumption:", self.energy_consumption,
                 "Duration:", self.duration,
                 "Power meter:", self.power_meter,
                 "Phone:", self.device_model)

    def __repr__(self):
        """Get representation of the measurement."""
        return (
            "Measurement("
            "use_case={use_case!r}, "
            "energy_consumption={energy_consumption!r}, "
            "duration={duration!r}, "
            "power_meter={power_meter!r}, "
            "device_model={device_model!r},"
            "success={success!r}"
            ")".format(**self.__dict__)
        )

    def __float__(self):
        """Convert measurement to float using energy consumption."""
        return self.energy_consumption

    @classmethod
    def clear_database(cls):
        """Clear database. Deletes CSV data file."""
        try:
            os.remove(cls.csv_storage)
        except OSError:
            pass

    @classmethod
    def _get_unique_from_column(cls, column_index):
        """Get unique values of the given column."""
        with open(cls.csv_storage, 'rt') as csvfile:
            csv_reader = csv.reader(csvfile)
            return {row[column_index] for row in csv_reader}

    @classmethod
    def get_unique_apps(cls):
        """Get all unique apps existing in the database.

        Returns:
            List of unique apps.

        """
        return cls._get_unique_from_column(cls.COLUMN_APP_PKG)

    @classmethod
    def get_unique_use_cases(cls, measurements):
        """Get all unique use cases.

        Returns:
            List of unique use cases.

        """
        return {
            measurement.use_case
            for measurement in measurements
        }

    @classmethod
    def get_all_entries_of_app(cls, app, use_case):
        """Get all entries that have a specific app and use case.

        If the use_case is None, all use_cases are retrieved.
        """
        with open(cls.csv_storage, 'rt') as csvfile:
            csv_reader = csv.reader(csvfile)
            return [
                Measurement(*row) for row in csv_reader
                if row[cls.COLUMN_USE_CASE] == use_case or use_case is None and
                row[cls.COLUMN_APP_PKG] == app
            ]

    @classmethod
    def get_entries_with_name_like(cls, name, measurements):
        """Get all measurements with a similar name to `name`."""
        return (
            measurement for measurement in measurements
            if name in measurement.use_case
        )

    @classmethod
    def get_entries_with_name(cls, name, measurements):
        """Get all measurements with a given name."""
        return (
            measurement for measurement in measurements
            if name == measurement.use_case
        )

    @classmethod # deleteme
    def mean_energy_consumption(cls, measurements):
        """Get mean energy consumption from a set of measurements."""
        len_measurements = len(measurements)
        if len_measurements == 0:
            raise Exception("Empty sample.")

        energy_consumptions = [
            float(measurement.energy_consumption)
            for measurement in measurements
        ]
        return sum(energy_consumptions) / len_measurements

    @classmethod
    def describe(cls, measurements):
        """Descriptive statistics for a set of measurements.

        Get descriptive statistics for time and energy
        consumption for a set of measurements.

        Returns:
            Tuple of Energy consumption mean, std, Duration mean, std.

        """
        len_measurements = len(measurements)
        if len_measurements == 0:
            return

        energy_consumptions = [
            float(measurement.energy_consumption)
            for measurement in measurements
        ]
        energy_consumption_mean = sum(energy_consumptions) / len_measurements
        energy_consumption_std = numpy.std(energy_consumptions)

        durations = [
            float(measurement.duration)
            for measurement in measurements
        ]
        duration_mean = sum(durations) / len_measurements
        duration_std = numpy.std(durations)
        return (
            energy_consumption_mean,
            energy_consumption_std,
            duration_mean,
            duration_std,
        )

    @classmethod
    def describe_app_use_case(cls, app, use_case):
        """Descriptive statistics for a stored App use case.

        Get descriptive statistics for time and energy
        consumption of an application use case.

        Args:
            app (string): Application package.
            use_case (string): Name of the use case

        Returns:
            Tuple of Energy consumption mean, std, Duration mean, std.

        """
        measurements = cls.get_all_entries_of_app(app, use_case)
        return cls.describe(measurements)


    @classmethod
    def get_energy_ranking(cls):
        """Ranking of the energy consumption of all apps.

        Get apps aggregated and sorted by mean energy consumption.

        Returns:
            OrderedDict with key=app_pkg and value=energy_consumption

        """
        with open(cls.csv_storage, 'rt') as csvfile:
            csv_reader = csv.reader(csvfile)

            data = [Measurement(*row) for row in csv_reader]
            data = sorted(data, key=lambda msrmnt: msrmnt.app_pkg)
            grouped_data = {
                k: Measurement.mean_energy_consumption(list(group))
                for (k, group) in groupby(
                    data,
                    key=lambda msrmnt: msrmnt.app_pkg
                )
            }
            sorted_data = OrderedDict(sorted(
                list(grouped_data.items()),
                key=itemgetter(1)
            ))
            return sorted_data

    @classmethod
    def get_position_in_ranking(cls, measurements):
        """Get the position in ranking of a given sample of measurements."""
        energy_ranking = cls.get_energy_ranking()
        consumptions = list(energy_ranking.values())
        energy_consumption = cls.mean_energy_consumption(measurements)
        return (
            bisect.bisect_left(consumptions, energy_consumption)+1,
            len(consumptions)
        )



"""Command Line Interface for Physalia.

Example:
        `$ python cli.py --serial 12886 -V 3.8  "sleep 5"`
"""

# pylint: disable=no-value-for-parameter
# pylint: disable=missing-docstring

import sys
import subprocess
import click
from physalia.energy_profiler import AndroidUseCase
from physalia.power_meters import MonsoonPowerMeter, EmulatedPowerMeter


@click.command()
@click.option('--count', default=1, type=click.IntRange(min=1),
              help='Number of measurement repetitions.')
@click.option('--power_meter', default='monsoon', type=click.Choice(['monsoon', 'dumb']),
              help="Which power meter to use.")
@click.option('-V', '--voltage', type=click.FLOAT,
              help="Set output voltage for Monsoon.")
@click.option('-s', '--serial', default=None, type=click.INT,
              help="Monsoon's serial number.")
@click.argument('exec_expression')
def tool(count, power_meter, voltage, serial, exec_expression):
    """Measure energy consumption while running a bash expression.

    Example:
        physalia --serial 12886 -V 3.8  "sleep 5"
    """
    physalia_power_meter = None
    if power_meter == 'monsoon':
        if not (voltage and serial):
            click.secho('Error: Monsoon requires to set voltage and serial number.', fg='red')
            sys.exit(-1)
        else:
            physalia_power_meter = MonsoonPowerMeter(voltage=voltage, serial=serial)
    elif power_meter == 'dumb':
        physalia_power_meter = EmulatedPowerMeter()

    def run(_):
        subprocess.check_output(exec_expression, shell=True)
    use_case = AndroidUseCase('physalia-cli', None, 'na', 'na', run=run)

    use_case.profile(
        power_meter=physalia_power_meter,
        verbose=True,
        retry_limit=3,
        count=count
    )


if __name__ == '__main__':
    tool()



"""Logic to asynchronously profile energy using Monsoon.

This is based on the original code by Google, which is
licensed under the Apache License, Version 2.0 .
"""

from threading import Thread
from Monsoon.sampleEngine import triggers

# pylint: disable=protected-access

class MonsoonReader(Thread):
    """`Thread` subclass to asynchronously control monsoon measurements."""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, monsoon_engine):
        super(MonsoonReader, self).__init__()
        self.monsoon_engine = monsoon_engine

    def prepare(self):
        """Prepare monsoon to start measuring."""
        # Make sure state is normal
        pass

    def run(self):
        """Start measuring."""
        self.monsoon_engine.startSampling(triggers.SAMPLECOUNT_INFINITE)

    def stop(self):
        """Stop measuring."""
        self.monsoon_engine._SampleEngine__stopTriggerSet = True
        self.join()



"""Asserts to use in energy tests."""

from physalia.models import Measurement

def consumption_below(sample, energy_consumption_baseline):
    """Test for energy consumption lower than a given value in Joules (avg).

    Args:
        sample (list of Measurement): sample of measurements
        energy_consumption (number): baseline energy consumption in Joules.
    """
    energy_consumption_mean = Measurement.mean_energy_consumption(sample)
    assert energy_consumption_mean < energy_consumption_baseline


def consumption_lower_than_app(sample, app, use_case=None):
    """Test that a given sample spends less energy than a known app.

    Args:
        sample (list of Measurement): sample of measurements
        app (string): identifier/package of the app to be compared
        use_case (string): select only data from a given use case
    """
    baseline_measurements = Measurement.get_all_entries_of_app(app, use_case)
    baseline_consumption = Measurement.mean_energy_consumption(
        baseline_measurements
    )
    consumption_below(sample, baseline_consumption)

def top_percentile(sample, nth):
    """Test that a given sample is in the top nth percentile.

    Args:
        sample (list of Measurement): sample of measurements
        nth (number): percentage of the position in which the sample should fit
        app (string): identifier of the application within the sample should be compared
        use_case (string: identifier of the use case used to create the ranking
    """
    position, total = Measurement.get_position_in_ranking(sample)
    percentile_position = float(position)/total*100
    assert percentile_position <= nth,\
           ("Given sample is not on {:.1f}% top percentile "
            "(Position: {:.1f}%)".format(
                nth,
                percentile_position
            ))



""" Test Models module
"""

import unittest
from physalia.models import Measurement
from physalia.fixtures.models import create_measurement
from physalia.fixtures.models import create_random_sample

# pylint: disable=missing-docstring

class TestMeasurement(unittest.TestCase):
    TEST_CSV_STORAGE = "./test_models_db.csv"

    def setUp(self):
        Measurement.csv_storage = self.TEST_CSV_STORAGE
        self.addCleanup(Measurement.clear_database)

    def test_persist(self):
        measurement = create_measurement()
        measurement.persist()
        with open(self.TEST_CSV_STORAGE, 'r') as file_desc:
            content = file_desc.read()
        self.assertIn(
            """1485634263.096069,login,com.package,1.0.0,Nexus 5X,2.0,30.0,NA,True""",
            content
        )

    def test_get_unique_apps(self):
        for _ in range(10):
            measurement = create_measurement(app_pkg="com.test.one")
            measurement.persist()
            measurement = create_measurement(app_pkg="com.test.two")
            measurement.persist()
            measurement = create_measurement(app_pkg="com.test.three")
            measurement.persist()
        self.assertEqual(
            set(Measurement.get_unique_apps()),
            {"com.test.one", "com.test.two", "com.test.three"}
        )

    def test_get_unique_use_cases(self):
        measurements = []
        for _ in range(10):
            measurements.append(create_measurement(use_case="one"))
            measurements.append(create_measurement(use_case="two"))
            measurements.append(create_measurement(use_case="three"))
        self.assertEqual(
            Measurement.get_unique_use_cases(measurements),
            {"one", "two", "three"}
        )

    def test_describe_app_use_case(self):
        for i in range(10):
            measurement = create_measurement(
                use_case="login",
                duration=i,
                energy_consumption=i * 2
            )
            measurement.persist()
        real_stats = Measurement.describe_app_use_case(
            measurement.app_pkg,
            measurement.use_case
        )
        expected_stats = (9, 5.745, 4.5, 2.872)
        # pairwise assertion:
        for (first, second) in zip(real_stats, expected_stats):
            self.assertAlmostEqual(first, second, places=3)


    def test_get_energy_ranking(self):
        sample = (
            create_random_sample(10, 1, app_pkg="com.app1") +
            create_random_sample(11, 1, app_pkg="com.app2") +
            create_random_sample(12, 1, app_pkg="com.app3") +
            create_random_sample(13, 1, app_pkg="com.app4") +
            create_random_sample(14, 1, app_pkg="com.app5") +
            create_random_sample(15, 1, app_pkg="com.app6")
        )
        for measurement in sample:
            measurement.persist()
        ranking = Measurement.get_energy_ranking()
        self.assertEqual(
            list(ranking.keys()),
            [
                "com.app1",
                "com.app2",
                "com.app3",
                "com.app4",
                "com.app5",
                "com.app6",
            ]
        )
        compare_sample = create_random_sample(12.5, 0.5, app_pkg="com.app7")
        self.assertEqual(
            Measurement.get_position_in_ranking(compare_sample),
            (4, 6)
        )



"""Provides functions for data analysis of Measurements."""

import sys
from string import Template
from operator import itemgetter
from collections import OrderedDict
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


from tabulate import tabulate
import tabulate as T
from scipy.stats import ttest_ind
from scipy.stats import f_oneway
# normality tests
from scipy.stats import normaltest, shapiro
import numpy as np

from physalia.utils.symbols import GREEK_ALPHABET

# pylint: disable=wrong-import-position
# Matplotlib is a special case
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from statsmodels.graphics.boxplots import violinplot as stats_violinplot


def violinplot(*samples, **options):
    """Create violin plot for a set of measurement samples."""
    names_dict = options.get("names_dict")
    title = options.get("title")
    sort = options.get("sort")
    millijoules = options.get("millijoules")

    consumptions = [np.array(sample, dtype='float') for sample in samples]
    if millijoules:
        for sample in consumptions:
            sample *= 1000
        unit = 'mJ'
    else:
        unit = 'J'

    if names_dict:
        labels = [
            sample and names_dict[sample[0].use_case]
            for sample in samples
        ]
    else:
        labels = [
            sample and sample[0].use_case.title().replace('_', ' ')
            for sample in samples
        ]

    if sort:
        labels, samples = zip(*sorted(zip(labels, samples)))

    stats_violinplot(consumptions, labels=labels, plot_opts={'label_rotation': 90})
    plt.gcf().subplots_adjust(bottom=0.3, left=0.1, right=0.999, top=0.99)
    axes = plt.gca()
    axes.set_ylim(bottom=0.0)
    axes.set_ylabel("Energy ({})".format(unit))
    axes.spines['right'].set_visible(False)
    axes.spines['left'].set_visible(False)
    axes.spines['top'].set_visible(False)
    axes.yaxis.grid(linestyle='dashed')
    axes.yaxis.set_ticks_position('none')

    if title:
        plt.title(title)
    if options.get('save_fig'):
        plt.savefig(options.get('save_fig'))
    if options.get('show_fig'):
        plt.show()


def samples_are_normal(*samples, **options):
    """Test whether each sample differs from a normal distribution.

    Use both Shapiro-Wilk test and D'Agostino and Pearson's test
    to test the null hypothesis that the sample is drawn from a
    normal distribution.

    Returns:
        List of tuples (is_normal(bool),(statistic(float),pvalue(float)).

    """
    alpha = options.get('alpha', 0.05)
    results = []
    for sample in samples:
        (_, shapiro_pvalue) = shapiro_result = shapiro(sample)
        (_, normaltest_pvalue) = normaltest_result = normaltest(sample)
        results.append((
            not (normaltest_pvalue < alpha and shapiro_pvalue < alpha),
            shapiro_result,
            normaltest_result
        ))
    return results


def hypothesis_test(sample_a, sample_b):
    """Perform hypothesis test over two samples of measurements.

    Uses Welch's t-test to check whether energy consumption
    is different in the populations of samples a and b.

    Args:
        sample_a (list of Measurement): measurements of sample a
        sample_b (list of Measurement): measurements of sample b

    Returns:
        t (float): The calculated t-statistic
        prob (float): The two-tailed p-value

    """
    return ttest_ind(
        [measurement.energy_consumption for measurement in sample_a],
        [measurement.energy_consumption for measurement in sample_b],
        equal_var=False
    )

def _format_test_result(result):
    statistic, pvalue = result
    return u"(test={:.2f}, {})".format(statistic, _pvalue_to_str(pvalue))

def pairwise_welchs_ttest(*samples, **options):
    """Perform pairwise Welch's t-test."""
    names = options.get("names")
    sort = options.get("sort")
    table_fmt = options.get("table_fmt", "grid")
    out = options.get("out", sys.stdout)

    if not names:
        names = [
            sample and sample[0].use_case.title().replace('_', ' ')
            for sample in samples
        ]

    if sort:
        names, samples = zip(*sorted(zip(names, samples)))

    samples = [np.array(sample, dtype='float') for sample in samples]
    len_samples = len(samples)
    table = list()
    for index, sample_one in enumerate(samples):
        row = list()
        for sample_two in samples[:index]:
            row.append(_format_test_result(ttest_ind(
                sample_one, sample_two,
                equal_var=False
            )))
        row.extend(["--"]*(len_samples-index))
        table.append(row)
    out.write(tabulate(table, headers=names, showindex=names, tablefmt=table_fmt))
    out.write("\n")

def fancy_hypothesis_test(sample_a, sample_b,
                          name_a, name_b, out=sys.stdout):
    """Perform and describe hypothesis testing of 2 samples.

    Args:
        sample_a (list of Measurement): measurements of sample a
        sample_b (list of Measurement): measurements of sample b
        sample_a (String): population name of sample a
        sample_b (String): population name of sample b
        out (file): data stream for output

    """
    alpha = 0.05
    test, pvalue = hypothesis_test(sample_a, sample_b)
    rejected_null_h = alpha <= pvalue
    out.write(
        Template(
            "Hypothesis testing:\n"
            "\t$H0: $mu {name_a} = $mu {name_b}.\n"
            "\t$H1: $mu {name_a} $neq $mu {name_b}.\n"
            "\n"
        ).substitute(GREEK_ALPHABET).format(name_a=name_a,
                                            name_b=name_b)
    )
    out.write(
        u"Applying Welch's t-test with {alpha_letter}=0.05, the null"
        u" hypothesis is{negate} rejected (p-value={pvalue}).\n".format(
            negate=" not" if rejected_null_h else "",
            pvalue="<0.001" if pvalue < 0.001 else "{:.3f}".format(pvalue),
            alpha_letter=GREEK_ALPHABET['alpha']
        ))

    if rejected_null_h:
        out.write("Thus, it was not possible to find evidence that"
                  " the means of populations {name_a} and {name_b}"
                  " are different.\n".format(name_a=name_a,
                                             name_b=name_b))
    else:
        out.write("Thus, one can say that the means of populations"
                  " \"{name_a}\" and \"{name_b}\" are different.\n"
                  "".format(name_a=name_a, name_b=name_b))
    return (test, pvalue)

def smart_hypothesis_testing(*samples, **options):
    """Do a smart hypothesis testing."""
    fancy = options.get('fancy', True)
    out = options.get('out', sys.stdout)
    alpha = options.get('alpha', 0.05)
    equal_var = options.get('equal_var', True)
    latex = options.get('latex', True)

    samples = [np.array(sample, dtype='float') for sample in samples]
    len_samples = len(samples)
    out_buffer = StringIO()

    normality_results = samples_are_normal(*samples)
    if all(map(itemgetter(0), normality_results)):
        # all our samples are normal
        if equal_var:
            if fancy:
                out_buffer.write(Template(
                    u"Hypothesis testing:\n\n"
                    "\t$H0: ${mu}1 = ${mu}2{ellipsis} = $mu{len_samples}. "
                    "The means for all groups are equal.\n"
                    "\t$H1: $exists a,b $elementof Samples: ${mu}a $neq ${mu}b. "
                    "At least two of the means are not equal.\n\n"
                    "The significance test one-way analysis of variance (ANOVA) "
                    "was used with a significance level of $alpha={alpha:.2f}.\n"
                    "This test requires that the following "
                    "assumptions are satisfied:\n\n"
                    "1. Samples are independent.\n"
                    "2. Samples are drawn from a normally distributed population.\n"
                    "3. All populations have equal standard deviation.\n\n"
                    "For the assumption of normal distribution two tests were "
                    "performed ($alpha={alpha}): Shapiro Wilk's test "
                    "and D'Agostino and Pearson's test.\n"
                    "None of these tests reject the null hypothesis with "
                    "significance level of $alpha={alpha}, thus it is assumed that data "
                    "follows a normal distribution.\n\n"
                    "").substitute(GREEK_ALPHABET).format(
                        ellipsis=" = ..." if len_samples > 3 else "",
                        **locals()
                    ))
            statistic, pvalue = f_oneway(*samples)
            if fancy:
                if pvalue < alpha:
                    out_buffer.write(
                        u"One can say that samples come from populations "
                        "with different means, since ANOVA rejects the "
                        "null hypothesis "
                        "(statistic={statistic:.2f}, {pvalue_str}).\n"
                        "".format(pvalue_str=_pvalue_to_str(pvalue), **locals())
                    )
                else:
                    out_buffer.write(
                        u"Thus, it was not possible to find evidence that"
                        " the means of populations are different "
                        "(statistic={statistic:.2f},{rho}={pvalue:.2f}).\n"
                        "".format(**locals())
                    )
            _flush_output(out, out_buffer, latex)
            return statistic, pvalue, f_oneway

def _pvalue_to_str(pvalue):
    pvalue_str = "<0.001" if pvalue < 0.001 else "={:.3f}".format(pvalue)
    return u"{}{}".format('p', pvalue_str)

def _flush_output(out, out_buffer, convert_to_latex):
    output = out_buffer.getvalue()
    out_buffer.close()
    if convert_to_latex:
        from pylatexenc.latexencode import utf8tolatex
        output = (
            "\\documentclass{article}\\begin{document}"
            "\\section{Physalia Hypothesis Test}\n" + utf8tolatex(output)
        ).replace(
            GREEK_ALPHABET["H0"], "$H_0$"
        ).replace(
            GREEK_ALPHABET["H1"], "$H_1$"
        ).replace(
            "<", "\\ensuremath{<}"
        ).replace(
            "\n", "\n\n"
        ).replace(
            "1.", '\\begin{enumerate}\n\\item '
        ).replace(
            "2.", '\\item '
        ).replace(
            "3. All populations have equal standard deviation.",
            '\\item All populations have equal standard deviation.\n\\end{enumerate}'
        ) + "\\end{document}"
    out.write(output)

def describe(*samples, **options):
    """Create table with statistic summary of samples.

    Args:
        loop_count
        names
        out
        table_fmt
        float_fmt
        show_ranking
        mili_joules
    """
    # pylint: disable=too-many-locals

    loop_count = options.get("loop_count")
    names = list(options.get("names"))
    out = options.get('out', sys.stdout)
    table_fmt = options.get("table_fmt", "grid")
    float_fmt = options.get("float_fmt", "")
    show_ranking = options.get("ranking")
    mili_joules = options.get("mili_joules")

    consumption_samples = [np.array(sample, dtype='float') for sample in samples]
    if mili_joules:
        for sample in consumption_samples:
            sample *= 1000
        unit = 'mJ'
    else:
        unit = 'J'
    samples_means = np.array([np.mean(sample) for sample in consumption_samples])
    if show_ranking:
        order = samples_means.argsort()
        ranking = order.argsort()

    table = list()
    for index, sample in enumerate(consumption_samples):
        mean = np.mean(sample)
        row = OrderedDict((
            ("N", len(sample)),
            ("$\\bar{{x}}$ ({})".format(unit), mean),
            ("$s$", np.std(sample)),
        ))
        if loop_count:
            #row["Iter."] = loop_count
            row["Single ({})".format(unit)] = mean/loop_count
        if show_ranking:
            row["Rank"] = int(ranking[index]+1)
            if row["Rank"] == 1 and table_fmt == 'latex':
                names[index] = "\\textbf{"+names[index]+"}"
        table.append(row)
    old_escape_rules = T.LATEX_ESCAPE_RULES
    T.LATEX_ESCAPE_RULES = {}
    out.write(T.tabulate(
        table,
        headers='keys', showindex=names,
        tablefmt=table_fmt, floatfmt=float_fmt
    ))
    T.LATEX_ESCAPE_RULES = old_escape_rules
    out.write("\n")
    return table



"""Example of using Physalia to profile F-Droid."""

from time import sleep
import subprocess
from samples.utils import AndroidViewClientUseCase

# pylint: disable=invalid-name
# this is just a script -- uppercase variables would look odd

def get_running_pkg():
    """Get Running App's package."""
    return subprocess.check_output(
        "adb shell dumpsys window windows |"
        " grep mCurrentFocus | cut -d'/' -f1 |"
        " rev | cut -d' ' -f1 | rev",
        shell=True
    ).strip()

def run_open_app(use_case):
    """Simply open app and wait 10 seconds."""
    use_case.open_app()
    sleep(10)

def cleanup(use_case):
    """Close the app and wait a second."""
    use_case.kill_app()
    sleep(1)

# Youtube Music

youtube_open_app_use_case = AndroidViewClientUseCase(
    "OpenApp",
    "./apks/YouTube Music_v1.62.7_apkpure.com.apk",
    "com.google.android.apps.youtube.music",
    "1.62.7",
    run_open_app,
    prepare=None,
    cleanup=cleanup
)

youtube_open_app_use_case.prepare_apk()
youtube_open_app_use_case.run()
youtube_open_app_use_case.profile_and_persist()

# Spotify

spotify_open_app_use_case = AndroidViewClientUseCase(
    "OpenApp",
    "./apks/Spotify Music_v7.5.0.1076_apkpure.com.apk",
    "com.spotify.music",
    "7.5.0.1076",
    run_open_app,
    prepare=None,
    cleanup=cleanup
)

spotify_open_app_use_case.prepare_apk()
spotify_open_app_use_case.run()
spotify_open_app_use_case.profile_and_persist()

# Google Play

google_play_open_app_use_case = AndroidViewClientUseCase(
    "OpenApp",
    "./apks/Google Play Music_v7.3.4313-1.M.3679657_apkpure.com.apk",
    "com.google.android.music",
    "7.3.4313.M.3679657",
    run_open_app,
    prepare=None,
    cleanup=cleanup
)

google_play_open_app_use_case.prepare_apk()
google_play_open_app_use_case.run()
google_play_open_app_use_case.profile_and_persist()

# SoundCloud

soundcloud_open_app_use_case = AndroidViewClientUseCase(
    "OpenApp",
    "./apks/SoundCloud Music Audio_v2017.01.24-release_apkpure.com.apk",
    "com.soundcloud.android",
    "2017.01.24-release",
    run_open_app,
    prepare=None,
    cleanup=cleanup
)

soundcloud_open_app_use_case.prepare_apk()
soundcloud_open_app_use_case.run()
soundcloud_open_app_use_case.profile_and_persist()

# Deezer

deezer_open_app_use_case = AndroidViewClientUseCase(
    "OpenApp",
    "./apks/Deezer Music Song Streaming_v5.4.0.88_apkpure.com.apk",
    "deezer.android.app",
    "5.4.0.88",
    run_open_app,
    prepare=None,
    cleanup=cleanup
)

deezer_open_app_use_case.prepare_apk()
deezer_open_app_use_case.run()
deezer_open_app_use_case.profile_and_persist()

# Apple Music

apple_music_open_app_use_case = AndroidViewClientUseCase(
    "OpenApp",
    "./apks/Apple Music_v1.2.0_apkpure.com.apk",
    "com.apple.android.music",
    "1.2.0",
    run_open_app,
    prepare=None,
    cleanup=cleanup
)

apple_music_open_app_use_case.prepare_apk()
apple_music_open_app_use_case.run()
apple_music_open_app_use_case.profile_and_persist()



"""Module with util functions to use with Monsoon."""

import usb.core
from math import isclose
from Monsoon import Operations
import click

def _device_matcher(device):
    """Check if device is monsoon.

    Method copied from official Monsoon API.
    """
    # pylint: disable=bare-except
    try:
        return device.idVendor == 0x2AB9 and device.idProduct == 0x0001
    except:
        return False

def is_monsoon_available():
    """Check wheter there is a Monsoon connnected."""
    device = usb.core.find(custom_match=_device_matcher)
    return device is not None

def get_voltage(monsoon):
    """Get voltage in the output of Monsoon."""
    read_voltage = monsoon.Protocol.getValue(Operations.OpCodes.setMainVoltage,1)
    return read_voltage / Operations.Conversion.FLOAT_TO_INT

def set_voltage_if_different(monsoon, voltage):
    """Set monsoon voltage only if it is different."""
    previous_voltage = get_voltage(monsoon)
    if not isclose(previous_voltage, voltage, abs_tol=0.0001):
        click.secho('setting voltage {} (previous was {})'.format(voltage, previous_voltage), fg='red')
        monsoon.setVout(voltage)



"""Models to interact with different power meters."""

import abc
import time
import warnings

import click
from Monsoon import Operations as operations
from Monsoon.sampleEngine import SampleEngine
from Monsoon import LVPM, HVPM
import Monsoon.pmapi as pmapi

from physalia.third_party import monsoon_async
from physalia.utils import android
from physalia.utils.monsoon import set_voltage_if_different

import numpy as np


class PowerMeter(object):
    """Abstract class for interaction with a power monitor."""

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def start(self):
        """Start measuring energy consumption."""
        return

    @abc.abstractmethod
    def stop(self):
        """Stop measuring energy consumption.

        Returns:
            tuple: energy consumption in Joules; duration; error flag.

        """
        return

    def reinit(self):
        """Reinitialize power meter upon unexpected behavior."""
        pass


class EmulatedPowerMeter(PowerMeter):
    """PowerMeter implementation to emulate a power monitor."""

    def __init__(self):  # noqa: D102,D107
        self.start_time = None

    def start(self):
        """Start measuring energy consumption."""
        self.start_time = time.time()

    def stop(self):
        """Stop measuring energy consumption.

        Returns:
            tuple: energy consumption in Joules; duration; error flag.

        """
        duration = time.time() - self.start_time
        energy_consumption = duration
        return energy_consumption, duration, False

    def __str__(self):
        """Return the name of this power meter."""
        return "Emulated"

class MonsoonPowerMeter(PowerMeter):
    """PowerMeter implementation for Monsoon LVPM.

    Make sure the Android device has Passlock disabled.
    Your server and device have to be connected to the same network.
    """

    def __init__(self, voltage=3.8, serial=None):  # noqa: D102,D107
        self.monsoon = None
        self.serial = serial
        self.voltage = voltage
        self.monsoon_reader = None
        self.monsoon_data = None
        self.engine = None
        self.setup_monsoon()

        click.secho(
            "Monsoon is ready.",
            fg='green'
        )
        if not android.is_android_device_available():
            click.secho(
                "You can now turn the phone on.",
                fg='blue'
            )
        for i in range(180):
            time.sleep(1)
            if android.is_android_device_available():
                time.sleep(2)
                click.secho(
                    "Found a {}!".format(android.get_device_model()),
                    fg='green'
                )
                break
            if i%15 == 0:
                click.secho(
                    "Waiting for an Android device...",
                    fg='blue'
                )
            if i == 180:
                raise Exception("Could not find device.")
        android.connect_adb_through_wifi()
        self.monsoon_usb_enabled(False)
        if android.is_locked():
            click.secho(
                "Device seems to be locked. "
                "Disabling Passlock is recommended!",
                fg='yellow'
            )

    def reinit(self):
        """Reinitialize power meter upon unexpected behavior."""
        warnings.warn(
            "reinit is deprecated and does nothing",
            DeprecationWarning
        )

    def setup_monsoon(self):
        """Set up monsoon.

        Args:
            voltage: Voltage output of the power monitor.
            serial: serial number of the power monitor.
        """
        click.secho(
            "Setting up Monsoon {} with {}V...".format(
                self.serial, self.voltage
            ),
            fg='blue'
        )
        self.monsoon = LVPM.Monsoon()
        self.monsoon.setup_usb(self.serial)
        set_voltage_if_different(self.monsoon, self.voltage)
        self.engine = SampleEngine(self.monsoon)
        self.engine.ConsoleOutput(False)

        if android.is_android_device_available():
            android.reconnect_adb_through_usb()
        self.monsoon_usb_enabled(True)

    def monsoon_usb_enabled(self, enabled):
        """Enable/disable monsoon's usb port."""
        # pylint: disable=too-many-function-args
        # something is conflicting with timeout_decorator
        self.monsoon.setUSBPassthroughMode(
            {
                True:operations.USB_Passthrough.On,
                False:operations.USB_Passthrough.Off,
            }[enabled]
        )

    def start(self):
        """Start measuring energy consumption."""
        self.monsoon_reader = monsoon_async.MonsoonReader(
            self.engine,
        )
        self.monsoon_reader.start()

    def stop(self):
        """Stop measuring."""
        self.monsoon_reader.stop()
        samples = self.engine.getSamples()
        if len(samples) == 3:
            timestamps = samples[0]
            currents = samples[1]
            if timestamps:
                sample_hz = 50000
                delta_time = 1/sample_hz
                time_deltas = [j-i for i, j in zip(timestamps[:-1], timestamps[1:])]
                
                energy_consumption = sum(np.array(currents[:-1])*np.array(time_deltas))/1000
                # energy_consumption = sum(currents)*delta_time
                duration = timestamps[-1]
                return energy_consumption, duration, False
        return None, None, True

    def __str__(self):
        """Return the name of this power meter."""
        return "Monsoon"

class MonsoonHVPMPowerMeter(MonsoonPowerMeter):
    def setup_monsoon(self):
        """Set up monsoon HVPM.

        Args:
            voltage: Voltage output of the power monitor.
            serial: serial number of the power monitor.
        """
        click.secho(
            "Setting up Monsoon {} with {}V...".format(
                self.serial, self.voltage
            ),
            fg='blue'
        )
        self.monsoon = HVPM.Monsoon()
        self.monsoon.setup_usb(self.serial, pmapi.USB_protocol())
        set_voltage_if_different(self.monsoon, self.voltage)
        self.engine = SampleEngine(self.monsoon)
        self.engine.ConsoleOutput(False)

        if android.is_android_device_available():
            android.reconnect_adb_through_usb()
        self.monsoon_usb_enabled(True)
    



"""Module with util functions to control Android devices."""

import subprocess
import re

from whichcraft import which
import click

def set_charging_enabled(enabled, serialno=None):
    """Enable or disable charging the device."""
    if serialno:
        command = (
            "adb shell -s {serialno} dumpsys battery set ac {enabled};"
            "adb shell -s {serialno} dumpsys battery set usb {enabled};"
        ).format(serialno=serialno, enabled=int(enabled))
    else:
        command = (
            "adb shell dumpsys battery set ac {enabled};"
            "adb shell dumpsys battery set usb {enabled}"
        ).format(enabled=int(enabled))

    subprocess.check_output(
        command,
        shell=True
    )

def prevent_device_from_sleep(enabled):
    """Prevent device from sleep while usb connected."""
    comand = "adb shell svc power stayon {}".format(
        {True: 'usb', False: 'false'}[enabled]
    )
    subprocess.check_output(
        comand,
        shell=True
    )

def is_screen_on():
    """Check whether the screen is on."""
    try:
        subprocess.check_output(
            "adb shell dumpsys input_method | grep mInteractive=true",
            shell=True
        )
        return True
    except subprocess.CalledProcessError:
        pass
    try:
        subprocess.check_output(
            'adb shell dumpsys power | grep "Display Power: state=ON"',
            shell=True
        )
        return True
    except subprocess.CalledProcessError:
        pass
    return False

def is_locked():
    """Check whether device is locked."""
    try:
        output = subprocess.check_output(
            "adb shell service call trust 7",
            shell=True,
            universal_newlines=True
        )
        match = re.search(r"Parcel\(00000000 00000001", output)
        return match is not None
    except subprocess.CalledProcessError as e:
        click.secho('Warning: {}'.format(e), fg='yellow')
        return True

def wakeup():
    """Wake up device."""
    if not is_screen_on():
        subprocess.check_output(
            "adb shell input keyevent 26",
            shell=True
        )

def unlock(pincode):
    """Unlock device with the given PIN."""
    wakeup()
    comand = (
        "adb shell input keyevent 82"
        " && adb shell input text {}"
        " && adb shell input keyevent 66"
    ).format(pincode)
    subprocess.check_output(
        comand,
        shell=True
    )

def install_apk(apk):
    """Install apk.
    Accepts Downgrade, grants all requestd permissions,
    and reinstalls if app already exists.
    """
    subprocess.check_output(["adb", "install", "-d", "-g", "-r", apk])

def check_adb():
    """Check whether adb is available."""
    return which("adb") is not None

def is_android_device_available():
    """Check whether there is at least an available android devices."""
    if not check_adb():
        return False
    result = subprocess.check_output(
        "adb devices",
        shell=True,
        universal_newlines=True
    )
    devices = result.partition('\n')[2].replace('\n', '').split('\tdevice')
    devices = [device for device in devices if len(device) > 2]
    if not devices:
        return False
    try:
        result = subprocess.check_output(
            "adb shell getprop sys.boot_completed",
            shell=True,
            universal_newlines=True).strip()
        return result == "1"
    except subprocess.CalledProcessError:
        return False

def get_device_model(serialno=None):
    """Get the currently connected device model."""
    if serialno:
        command = ("adb shell -s {} "
                   "getprop ro.product.model").format(serialno)
    else:
        command = "adb shell getprop ro.product.model"
    try:
        return subprocess.check_output(
            command,
            shell=True,
            universal_newlines=True
        ).strip()
    except subprocess.CalledProcessError:
        return "N/A"

def connect_adb_through_wifi():
    """Configure `adb` through a wifi connection."""
    net_output = subprocess.check_output(
        "adb shell ip -f inet addr show wlan0",
        shell=True,
        universal_newlines=True
    )
    ip_address = re.search(r"inet \d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}", net_output).group()[5:]
    subprocess.check_output(
        "adb tcpip 5555",
        shell=True
    )
    subprocess.check_output(
        "adb connect {}".format(ip_address),
        shell=True
    )

def reconnect_adb_through_usb():
    """Connect adb back to USB while in wifi."""
    try:
        subprocess.check_output(
            "adb reconnect",
            shell=True
        )
    except subprocess.CalledProcessError:
        pass

def get_package_from_apk(apk_path):
    aapt = "$ANDROID_HOME/build-tools/27.0.0/aapt"
    result = subprocess.check_output(
        aapt + " dump badging {} | awk '/package/{{gsub(\"name=|'\"'\"'\",\"\");  print $2}}'".format(apk_path),
        shell=True
    )
    return str(result.strip(), 'utf-8')

def get_instrumentation_for_app(app_pkg, test_pkg=""):
    pattern = re.compile("instrumentation:(.*) ")
    output = subprocess.check_output(
        "adb shell pm list instrumentation | grep -i {}".format(app_pkg),
        shell=True
    )
    if type(output) is bytes:
        output = output.decode('utf-8')
    search = pattern.search(output)
    if search:
        return search.group(1)
    



"""Test energy_profiler module."""

import unittest

from physalia.energy_profiler import AndroidUseCase

# pylint: disable=missing-docstring

class TestEnergyProfiler(unittest.TestCase):

    def test_empty_android_use_case(self):
        # pylint: disable=no-self-use
        use_case = AndroidUseCase(
            name="Test",
            app_apk="no/path",
            app_pkg="no.package",
            app_version="0.0.0",
            run=None,
            prepare=None,
            cleanup=None
        )
        use_case.run()



"""Fixtures for models module."""

from physalia.models import Measurement
import numpy


def create_measurement(use_case='login',
                       app_pkg='com.package',
                       duration=2,
                       energy_consumption=30):
    """Fake data for measurement."""
    return Measurement(
        1485634263.096069,  # timestamp
        use_case,           # use_case
        app_pkg,            # application package
        '1.0.0',            # version
        'Nexus 5X',         # device model
        duration,           # duration
        energy_consumption  # energy consumption
    )

def create_random_sample(mean, std,
                         app_pkg='com.package',
                         use_case='login',
                         count=30, seed=1):
    """Create a sample of measurements."""
    # pylint: disable=too-many-arguments
    if seed is not None:
        numpy.random.seed(seed)
    energy_consumptions = numpy.random.normal(loc=mean,
                                              scale=std,
                                              size=count)
    return [
        create_measurement(
            energy_consumption=energy_consumptions[i],
            app_pkg=app_pkg,
            use_case=use_case
        )
        for i in range(count)
    ]

def create_random_samples(count=30, seed=1):
    """Create a sample of measurements."""
    if seed is not None:
        numpy.random.seed(seed)
    sample_a = create_random_sample(10.0, 1.0, count=count, seed=None)
    sample_b = create_random_sample(12.0, 1.0, count=count, seed=None)
    return sample_a, sample_b



"""Test power_meters module."""

import time
import unittest

from physalia.power_meters import MonsoonPowerMeter
from physalia.utils.monsoon import is_monsoon_available

# pylint: disable=missing-docstring

class TestPowerMeters(unittest.TestCase):

    @unittest.skipUnless(is_monsoon_available(),
                         "Monsoon is required to run this test.")
    def test_collect_sample(self):
        desired_duration = 2
        power_meter = MonsoonPowerMeter(voltage=3.8, serial=12886)
        power_meter.start()
        time.sleep(desired_duration)
        energy_consumption, duration, error = power_meter.stop()
        self.assertFalse(error, "Power meter flagged an error.")
        self.assertAlmostEqual(duration, desired_duration, delta=1)
        self.assertGreater(energy_consumption, 0)

    @unittest.skipUnless(is_monsoon_available(),
                         "Monsoon is required to run this test.")
    def test_collect_tiny_sample(self):
        """Test measurements with less than a second.

        There was originally an error in which samples with less than 1
        second would not return any measurements.
        Still, such short measurements are not recommended.
        The first 0.2 seconds are not returning any data.
        """

        desired_duration = 0.3
        power_meter = MonsoonPowerMeter(voltage=3.8, serial=12886)
        power_meter.start()
        time.sleep(desired_duration)
        energy_consumption, _, error = power_meter.stop()
        self.assertFalse(error, "Power meter flagged an error.")
        self.assertGreater(energy_consumption, 0)
        


"""Example that measures energy consumption of a mobile device for 5 secs."""

# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# reason: this is a very basic usage example

import time
from physalia.power_meters import MonsoonPowerMeter
from physalia.energy_profiler import AndroidUseCase

power_meter = MonsoonPowerMeter(voltage=3.8, serial=12886)

def run(_):
    time.sleep(5)
use_case = AndroidUseCase('physalia-simple-tutorial', None, 'na', 'na', run=run)

measurement = use_case.run(power_meter=power_meter)
print(measurement)



from physalia.power_meters import MonsoonPowerMeter
from physalia.energy_profiler import AndroidUseCase
from time import sleep


power_meter = MonsoonHVPMPowerMeter(voltage=3.8, serial=None)

def run(usecase):
	sleep(2) # some work
	
use_case = AndroidUseCase(
  'login',
  'path/to/apk',
  'com.test.app',
  '0.0',
  prepare=None,
  run=run,
  cleanup=None
)
measurement = use_case.run(power_meter=power_meter)
print(measurement)



"""Example of the usage of AndroidViewClientUseCase for
with F-Droid.
"""

from time import sleep
from samples.utils import AndroidViewClientUseCase

# pylint: disable=invalid-name
# this is just a script -- uppercase variables would look odd

def run_open_app(use_case):
    """Simply open F-droid."""
    use_case.open_app()
    sleep(2)
    use_case.refresh()
    use_case.view_client.findViewWithText(u'''What's New''')


def cleanup(use_case):
    """Close the app and wait a second."""
    use_case.kill_app()
    sleep(1)


open_app_use_case = AndroidViewClientUseCase(
    "OpenApp",
    "./fdroid.apk",
    "org.fdroid.fdroid",
    "0.01",
    run_open_app,
    prepare=None,
    cleanup=cleanup
)

open_app_use_case.prepare_apk()
open_app_use_case.profile_and_persist()


# View a listed app
def prepare(use_case):
    """Open app and wait until it loads."""
    use_case.open_app()
    use_case.refresh()
    use_case.wait_for_id("org.fdroid.fdroid:id/icon")


def run_view_listed_app(use_case):
    """Interaction of getting info about an application."""
    app_icon = use_case.view_client.findViewById("org.fdroid.fdroid:id/icon")
    app_icon.touch()
    use_case.wait_for_id("org.fdroid.fdroid:id/ll_description")


view_listed_app_use_case = AndroidViewClientUseCase(
    "ViewListedApp",
    "./fdroid.apk",
    "org.fdroid.fdroid",
    "0.01",
    run_view_listed_app,
    prepare=prepare,
    cleanup=cleanup
)

view_listed_app_use_case.profile_and_persist()

def run_search_for_app(use_case):
    """Interaction of searching for an application."""
    search = use_case.view_client.findViewById("org.fdroid.fdroid:id/action_search")
    search.touch()
    sleep(1)
    use_case.device.type("Frood")
    use_case.wait_for_text("Froody")

search_app_use_case = AndroidViewClientUseCase(
    "SearchApp",
    "./fdroid.apk",
    "org.fdroid.fdroid",
    "0.01",
    run_search_for_app,
    prepare=prepare,
    cleanup=cleanup
)

search_app_use_case.profile_and_persist()



"""Test power_meters module."""

import time
import unittest

import Monsoon.LVPM as LVPM

from physalia.utils.monsoon import is_monsoon_available, get_voltage

# pylint: disable=missing-docstring

class TestMonsoonUtils(unittest.TestCase):

    @unittest.skipUnless(is_monsoon_available(),
                         "Monsoon is required to run this test.")
    def test_get_voltage(self):
        voltage = 4.0
        monsoon = LVPM.Monsoon()
        monsoon.setup_usb()
        monsoon.setVout(voltage)
        self.assertEqual(voltage, get_voltage(monsoon))



"""Module with main classes for energy profiling."""

import time
import subprocess
import types
import click
from physalia.power_meters import EmulatedPowerMeter
from physalia.models import Measurement
import physalia.utils.android as android_utils
from physalia.exceptions import PhysaliaExecutionFailed


class AndroidUseCase(object):
    """Implementation of an Android use case.

    Args:
        name            name identifier of the use case
        app_apk         path of the app apk
        app_pkg         package
        power_meter     power meter to use for measurements
        app_version     version of the app
        run             method with Android interaction
        prepare         method to run before interaction
        cleanup         method to run after interaction

    """

    # pylint: disable=not-callable
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # Eight is reasonable in this case.

    default_power_meter = EmulatedPowerMeter()

    def __init__(self, name, app_apk, app_pkg, app_version,
                 run=None, prepare=None, cleanup=None):  # noqa: D102,D107
        self.name = name
        self.app_apk = app_apk
        self.app_pkg = app_pkg
        self.app_version = app_version
        self.notes = None
        if run:
            self._run = types.MethodType(run, self)
        if prepare:
            self._prepare = types.MethodType(prepare, self)
        if cleanup:
            self._cleanup = types.MethodType(cleanup, self)

    def _prepare(self):
        # pylint: disable=method-hidden
        pass

    def _cleanup(self):
        # pylint: disable=method-hidden
        pass

    def _run(self):
        # pylint: disable=method-hidden
        pass

    def prepare(self):
        """Prepare environment for running."""
        self._prepare()

    def cleanup(self):
        """Clean environment after running."""
        self._cleanup()

    def run(self, power_meter=default_power_meter, retry_limit=1):
        """Measure the routine stored in `_run`.

        Returns:
            Measurement: data collected from experiment

        """
        try:
            self.prepare()
            power_meter.start()
            success = self._run()
            energy_consumption, duration, error_flag = power_meter.stop()
            self.cleanup()
            if error_flag:
                raise PhysaliaExecutionFailed()
            
            measurement = Measurement(
                time.time(),
                self.name,
                self.app_pkg,
                self.app_version,
                android_utils.get_device_model(),
                duration,
                energy_consumption,
                str(power_meter),
                success is None or success,
                self.notes
            )
            
            return measurement
        except KeyboardInterrupt as error:
            raise error
        except BaseException as error:
            click.secho("Measurement {} has failed".format(self.name), fg='red')
            click.secho(str(error), fg='red')
            if retry_limit == 1:
                # click.secho("Waiting 5 minutes for Monsoon to recover...", fg='red')
                # time.sleep(5*60)
                # hack: reinit power_meter
                power_meter.reinit()

            if retry_limit > 0:
                click.secho("Retrying...", fg='yellow')
                return self.run(power_meter, retry_limit-1)
            printstack = getattr(error, "printStackTrace", None)
            if callable(printstack):
                printstack()
            raise error

    def profile(self, power_meter=default_power_meter,
                verbose=True, count=30, retry_limit=1,
                save_to_csv=None):
        """Run a batch of measurements.

        Args:
            power_meter     Power meter to use in measurements.
            verbose         Log activiy (default=True).
            count           Run experiment several times (default=30).
            retry_limit     Number of times to retry on error.
            save_to_csv     File name to store measurement.
        Returns: Set of measurements

        """
        results = []
        for i in range(count):
            result = self.run(power_meter=power_meter, retry_limit=retry_limit)
            if result:
                results.append(result)
                if save_to_csv:
                    result.save_to_csv(save_to_csv)
            else:
                click.secho("Error in execution {} of {}. Skipping.".format(i, self.name), fg="red")
        if verbose and results:
            click.secho("Energy consumption results for {}: "
                        "{:.3f} Joules (s = {:.3f}).\n"
                        "It took {:.1f} seconds (s = {:.1f})."
                        .format(self.app_pkg, *Measurement.describe(results)),
                        fg='green')
        return results

    def profile_and_persist(self, power_meter=default_power_meter,
                            verbose=True, count=30):
        """Measure a batch of measurements and save it."""
        results = self.profile(power_meter, verbose, count)
        for measurement in results:
            measurement.persist()
        return results

    def uninstall_app(self):
        """Uninstall app of the Android device."""
        click.secho("Uninstalling {}".format(self.app_pkg), fg='blue')
        subprocess.check_output(["adb", "uninstall", self.app_pkg])

    def install_app(self):
        """Install App."""
        click.secho("Installing {}".format(self.app_apk), fg='blue')
        android_utils.install_apk(self.app_apk)

    def prepare_apk(self):
        """Reinstall app in the Android device."""
        self.uninstall_app()
        self.install_app()

    def open_app(self):
        """Open app in the device."""
        click.secho("Opening app {}".format(self.app_pkg), fg='blue')
        subprocess.check_output(
            "adb shell monkey -p {} --pct-syskeys 0 1".format(self.app_pkg),
            shell=True
        )

    def kill_app(self):
        """Tell the device to kill the app of this use case."""
        click.secho("Killing app {}".format(self.app_pkg), fg='blue')
        subprocess.check_output(
            "adb shell am force-stop {}".format(self.app_pkg),
            shell=True
        )



"""Utils for Physalia samples"""

import sys
from com.dtmilano.android.viewclient import ViewClient
from physalia.energy_profiler import AndroidUseCase

class AndroidViewClientUseCase(AndroidUseCase):
    """`AndroidUseCase` to use with `AndroidViewClient`."""

    # pylint: disable=not-callable
    # pylint: disable=too-many-arguments
    # Eight is reasonable in this case.

    def __init__(self, name, app_apk, app_pkg, app_version,
                 run, prepare=None, cleanup=None):  # noqa: D102
        super(AndroidViewClientUseCase, self).__init__(
            name, app_apk, app_pkg, app_version,
            run, prepare, cleanup
        )
        self.device = None
        self.serialno = None
        self.view_client = None

    def start_view_client(self, force=False):
        """Setup `AndroidViewClient`.

        Args:
            force (boolean): force start even if it was previously done (default False).
        """
        # pylint: disable=attribute-defined-outside-init

        if self.view_client is None or force is True:
            original_argv = sys.argv
            sys.argv = original_argv[:1]
            kwargs1 = {'ignoreversioncheck': False,
                       'verbose': False,
                       'ignoresecuredevice': False}
            device, serialno = ViewClient.connectToDeviceOrExit(**kwargs1)
            kwargs2 = {'forceviewserveruse': False,
                       'useuiautomatorhelper': False,
                       'ignoreuiautomatorkilled': True,
                       'autodump': False,
                       'startviewserver': True,
                       'compresseddump': True}
            view_client = ViewClient(device, serialno, **kwargs2)
            sys.argv = original_argv
            self.device = device
            self.serialno = serialno
            self.view_client = view_client
        #always refresh
        self.refresh()

    def prepare(self):
        """Prepare environment for running.

        Setup Android View Client in order to run experiments.
        """
        self.start_view_client()
        self._prepare()

    def refresh(self):
        """Refresh `AndroidViewClient`."""
        while True:
            try:
                self.view_client.dump(window='-1')
            except RuntimeError:
                continue
            break

    def wait_for_id(self, view_id):
        """Refresh `AndroidViewClient` until view id is found."""
        view = self.view_client.findViewById(view_id)
        while view is None:
            self.refresh()
            view = self.view_client.findViewById(view_id)
        return view

    def wait_for_text(self, text):
        """Refresh `AndroidViewClient` until text is found."""
        view = self.view_client.findViewWithText(text)
        while view is None:
            self.refresh()
            view = self.view_client.findViewWithText(text)
        return view

    def wait_for_content_description(self, content_description):
        """Refresh `AndroidViewClient` until content description is found."""
        view = self.view_client.findViewWithContentDescription(
            content_description
        )
        while view is None:
            self.refresh()
            view = self.view_client.findViewWithContentDescription(
                content_description
            )
        return view



"""Test analytics module."""

import unittest
from tempfile import NamedTemporaryFile

from mock import patch, MagicMock

from physalia.analytics import violinplot
from physalia.fixtures.models import create_random_sample, create_random_samples
from physalia.analytics import hypothesis_test, fancy_hypothesis_test, smart_hypothesis_testing
from physalia.utils.symbols import GREEK_ALPHABET


# pylint: disable=missing-docstring

class TestAnalytics(unittest.TestCase):

    def test_violinplot(self):
        # pylint: disable=no-self-use
        sample_a = create_random_sample(10, 1, use_case='login_fb')
        sample_b = create_random_sample(20, 0.5, use_case='login_twitter')
        sample_c = create_random_sample(15, 3, use_case='login_google+')
        with NamedTemporaryFile(prefix="violinplot",
                                suffix='.png', delete=False) as tmp_file:
            violinplot(sample_a, sample_b, sample_c,
                       save_fig=tmp_file)

    def test_hypothesis_test(self):
        sample_a, sample_b = create_random_samples()
        _, pvalue = hypothesis_test(sample_a, sample_b)
        self.assertLess(pvalue, 0.05)

    def test_fancy_hypothesis_test(self):
        try:
            from StringIO import StringIO
        except ImportError:
            from io import StringIO
        from string import Template

        # sample_a, sample_b = create_random_samples()
        with patch('physalia.analytics.hypothesis_test',
                   MagicMock(return_value=(0, 0.0001))):
            out = StringIO()
            fancy_hypothesis_test(
                None, None,
                "login with email",
                "login with facebook",
                out=out
            )
            output = out.getvalue()

            self.assertEqual(
                output,
                Template(
                    "Hypothesis testing:\n"
                    "\t$H0: $mu login with email = $mu login with facebook.\n"
                    "\t$H1: $mu login with email $neq $mu login with facebook.\n"
                    "\n"
                    "Applying Welch's t-test with $alpha=0.05, the null"
                    " hypothesis is rejected (p-value=<0.001).\n"
                    "Thus, one can say that the means of populations"
                    " \"login with email\" and \"login with facebook\" are"
                    " different.\n"
                ).substitute(GREEK_ALPHABET)
            )
        with patch('physalia.analytics.hypothesis_test',
                   MagicMock(return_value=(0, 0.061231))):
            out = StringIO()
            fancy_hypothesis_test(
                None, None,
                "login with email",
                "login with facebook",
                out=out
            )
            output = out.getvalue()
            self.assertEqual(
                output,
                Template(
                    "Hypothesis testing:\n"
                    "\t$H0: $mu login with email = $mu login with facebook.\n"
                    "\t$H1: $mu login with email $neq $mu login with facebook.\n"
                    "\n"
                    "Applying Welch's t-test with $alpha=0.05, the null"
                    " hypothesis is not rejected (p-value=0.061).\n"
                    "Thus, it was not possible to find evidence that"
                    " the means of populations login with email and "
                    "login with facebook are different.\n"
                ).substitute(GREEK_ALPHABET)
            )

    def test_smart_hypothesis_testing(self):
        # pylint: disable=no-self-use
        sample_a = create_random_sample(10, 1, use_case='login_fb')
        sample_b = create_random_sample(20, 0.5, use_case='login_twitter')
        sample_c = create_random_sample(15, 3, use_case='login_google+')
        with NamedTemporaryFile(
            prefix="violinplot",
            suffix='.tex', delete=False
        ) as tmp_file, open(tmp_file.name, "w") as out:
            smart_hypothesis_testing(
                sample_a, sample_b, sample_c,
                fancy="True", alpha=0.05, equal_var=True,
                out=out
            )



"""Test Assert module."""

import unittest
from physalia import asserts
from physalia.fixtures.models import create_random_sample
from physalia.models import Measurement

# pylint: disable=missing-docstring

class TestAssert(unittest.TestCase):
    TEST_CSV_STORAGE = "./test_asserts_db.csv"

    def setUp(self):
        Measurement.csv_storage = self.TEST_CSV_STORAGE
        self.addCleanup(Measurement.clear_database)

    def test_consumption_below(self):
        sample = create_random_sample(10, 1)
        asserts.consumption_below(sample, 11)
        with self.assertRaises(Exception):
            asserts.consumption_below(sample, 9)

    def test_consumption_lower_than_app(self):
        sample_low_energy = create_random_sample(
            9, 1,
            app_pkg='com.sample',
            use_case='login'
        )
        sample_high_energy = create_random_sample(
            12, 1,
            app_pkg='com.sample',
            use_case='login'
        )
        existing_sample_one = create_random_sample(
            10, 1,
            app_pkg='com.persisted',
            use_case='login'
        )
        existing_sample_two = create_random_sample(
            11, 1,
            app_pkg='com.persisted',
            use_case='logout'
        )

        for measurement in existing_sample_one+existing_sample_two:
            measurement.persist()

        asserts.consumption_lower_than_app(
            sample_low_energy, "com.persisted"
        )
        asserts.consumption_lower_than_app(
            sample_low_energy, "com.persisted", "login"
        )
        with self.assertRaises(Exception):
            asserts.consumption_lower_than_app(
                sample_high_energy, "com.persisted"
            )
        with self.assertRaises(Exception):
            asserts.consumption_lower_than_app(
                sample_high_energy, "com.persisted", "login"
            )

    def test_top_percentile(self):
        sample = create_random_sample(
            11, 1,
            app_pkg='com.sample',
            use_case='login'
        )
        for i in range(100):
            existing_sample = create_random_sample(
                i, 1,
                app_pkg=('com.persisted.{}'.format(i)),
                use_case='login'
            )
            for measurement in existing_sample:
                measurement.persist()
        asserts.top_percentile(sample, 12)
        with self.assertRaises(Exception):
            asserts.top_percentile(sample, 11)
